#library(mvtnorm)
#library(gdata)
#library(dr)
library(rdetools)
#library(penalized)
library(sandwich)
library(parcor)

KM.matrix <- function(Z, kernel = "linear", weights = NULL, sigma = 1)
  {
  	n = nrow(Z); m = ncol(Z)
  	if (class(kernel) == "matrix"){
  	  	K = kernel
  	}else{
    	if (kernel == "linear"){K = Z%*%t(Z)}
    	if (kernel == "IBS") {
      		K = matrix(0, nrow = n, ncol = n)
			for (k in 1:m) {K = K + (2-abs(Z[,k] - VTM(Z[,k],n)))/(2*m)}}
    	if (kernel == "IBS.weighted") {
      		K = matrix(0,nrow = n, ncol = n)
      		if (is.nul(weights)) {qs = apply(Z == 2, 2, sum)/(2*n); weights = 1/sqrt(qs)}
      		for (k in 1:m) {
				K = K + ((2-abs(Z[,k] - VTM(Z[,k],n)))*weights[k])/(sum(weights))}}
    	if (kernel == "quadratic"){K = (Z%*%t(Z)+sigma)**2 }
		if (kernel == "RBF"){K = rbfkernel(Z, sigma = sigma)}
  	}
   K
  }

gen.data <- function(n.data, alpha0, beta0, h0, h1, g = c("linear", "logit"),sigma, rho, p, G.prob = .5, sigma_type = c("diag", "AR-1", "exhchangeable"),  X.G.dependent = FALSE){

    vcov <- diag(1, p); if(length(sigma)==1){sigma=diag(sigma,p)}
    if (sigma_type == "AR-1"){
        vcov <- sigma%*%rho^abs(row(vcov)-col(vcov))%*%sigma
    }
    if (sigma_type == "exchangeable"){
		vcov <- sigma%*%((1-rho)*diag(1,p,p)+rho)%*%sigma
    }
    G <- 2*rbinom(n.data,1,G.prob)-1
    X <- rmvnorm(n.data, mean = rep(0,p), sigma = vcov)

    if (g == "linear"){
        means <- alpha0 + beta0*G + h0(X) + h1(X)*(G+1)/2
        Y <- means + rnorm(n.data)*sd.fun(X)
    }

    if (g == "logit"){
        expit <- function(x) exp(x)/(1 + exp(x))
        means <- expit(alpha0 + beta0*G + h0(X) + h1(X)*(G+1)/2)
        Y <- rbinom(n.data,1,means)
    }
    return(cbind(Y, G, X))
}

power_sim_noPCA_new <- function(dat, n.data, p, g, kernel,kernpercent=99.9, nsample=1, sig_range){ 
  myfam = ifelse(g=="logit","binomial","gaussian"); if(g=="logit"){g.link=function(xx){exp(xx)/(1+exp(xx))}}else{g.link=function(xx){xx}}
  pvalue_list=matrix(0,nsample,length(sig_range))
  pvalue_KM=rep(0,nsample)
  for (j in 1:nsample){
  Y = dat[,1]; G = dat[,2]; X = dat[,c(-1,-2)]; n1 = sum(G==1); n0 = sum(G==-1); 
  BB = 1000; Vi.ptb = matrix(rnorm(n.data*BB),ncol=BB)
  Y1=Y[G==1]; Y0=Y[G==-1]; X1=X[G==1,]; X0=X[G==-1,]
  
  mu1.nul = mean(Y[G== 1]); fit1 = glm(Y~X,subset=G== 1,family=myfam); mu1.alt = NULL; x1bar = apply(X[G== 1,],2,mean)
  mu0.nul = mean(Y[G==-1]); fit0 = glm(Y~X,subset=G==-1,family=myfam); mu0.alt = NULL; x0bar = apply(X[G==-1,],2,mean)
  
  mu0.alt_ori = X0%*%adalasso(X0, Y0)$coefficients.adalasso; mu0.alt = mu0.alt_ori = c(g.link(glm(Y0~1,offset=mu0.alt_ori,family=myfam)$coef+mu0.alt_ori))
  mu1.alt_ori = X1%*%adalasso(X1, Y1)$coefficients.adalasso; mu1.alt = mu1.alt_ori = c(g.link(glm(Y1~1,offset=mu1.alt_ori,family=myfam)$coef+mu1.alt_ori))
  
  if (kernel=="linear"){
    kernmat = KM.matrix(Z=as.matrix(dat[,c(-1,-2)]),kernel=kernel,weights=NULL,sigma=sigma.tmp)/dim(dat)[1]
    mu.K = apply(kernmat,2,mean); kernmat = kernmat - mu.K - VTM(mu.K,n.data) + mean(mu.K); 
    Phi <- eigen(kernmat); mm = which.min(abs(round(cumsum(Phi$values)/sum(Phi$values)-(kernpercent/100),4))); ##print(Phi$values[1:mm])
    Phi = Phi$vectors[,1:mm]%*%diag(sqrt(Phi$values[1:mm])); 
    Shat = apply((Phi*(Y-mu1.nul))[G==1,],2,mean)-apply((Phi*(Y-mu0.nul))[G==-1,],2,mean)
    tmp.e1 = (Phi[G== 1,]-VTM(apply(Phi[G== 1,],2,mean),n1))*(Y[G==  1]-mu1.alt); #tmp.e1 = tmp.e1 - VTM(apply(tmp.e1,2,mean),nrow(tmp.e1))
    tmp.e0 = (Phi[G==-1,]-VTM(apply(Phi[G==-1,],2,mean),n0))*(Y[G== -1]-mu0.alt); #tmp.e0 = tmp.e0 - VTM(apply(tmp.e0,2,mean),nrow(tmp.e0))
    Shat.ptb = t(Vi.ptb[G== 1,])%*%tmp.e1/n1 -  t(Vi.ptb[G==-1,])%*%tmp.e0/n0;  ## B x mm matrix
    Qhat = sum(Shat^2); Qptb = apply(Shat.ptb^2,1,sum); pvalue.I = mean(Qptb > Qhat); pvalue_KM_original[j] = pvalue.I
    Var.Shat = var(Shat.ptb); invV.Shat = solve(diag(diag(Var.Shat)))
    Qhat = c(t(Shat)%*%invV.Shat%*%Shat); Qptb = apply(Shat.ptb*(Shat.ptb%*%invV.Shat),1,sum); 
    pvalue.Vinv = mean(Qptb > Qhat); pvalue_KM[j] = pvalue.Vinv
    S.out[j,]=c(Shat,rep(NA,p-length(Shat)),apply(Shat.ptb,2,sd),rep(NA,p-length(Shat)))
  }
  
  if ((kernel=="RBF")|(kernel=="quadratic")){
    accept =rep(1,length(sig_range)); rate.vec = mm.vec = rep(0,length(sig_range)); 
    sQ_ptb = pvalue_ptb=matrix(0,length(sig_range),BB); pvalue=rep(0,length(sig_range))
    for (kk in 1:length(sig_range)){
      sigma.tmp = sig_range[kk]
      kernmat = KM.matrix(Z=as.matrix(dat[,c(-1,-2)]),kernel=kernel,weights=NULL,sigma=sigma.tmp)/dim(dat)[1]; mu.K = apply(kernmat,2,mean)
      kernmat = kernmat - mu.K - VTM(mu.K,n.data) + mean(mu.K); Phi <- eigen(kernmat); Phi$values = pmax(Phi$values,0)
      mm = which.min(abs(cumsum(Phi$values)/sum(Phi$values)-(kernpercent/100)))
      rate.vec[kk] = rlm(log(Phi$values[1:mm])~log(1:mm))$coef[2]
      Phi = Phi$vectors[,1:mm]%*%diag(sqrt(Phi$values[1:mm]))
      Shat = apply((Phi*(Y-mu1.nul))[G==1,],2,mean)-apply((Phi*(Y-mu0.nul))[G==-1,],2,mean)
      Shat.ptb = t(Vi.ptb[G== 1,])%*%((Phi[G== 1,]-VTM(apply(Phi[G== 1,],2,mean),n1))*(Y[G== 1]-mu1.alt))/n1 -
        t(Vi.ptb[G==-1,])%*%((Phi[G==-1,]-VTM(apply(Phi[G==-1,],2,mean),n0))*(Y[G==-1]-mu0.alt))/n0;  ## B x mm matrix
      Qhat = sum(Shat^2); Qptb = apply(Shat.ptb^2,1,sum); pvalue[kk] = mean(Qptb > Qhat); pvalue_ptb[kk,] = 1-rank(Qptb)/length(Qptb)
    }
    null_dist=apply(pvalue_ptb[(accept==1),],2,min); obs_stat=min(pvalue[accept==1])
    pvalue_KM[j] = mean(obs_stat>null_dist); pvalue_list[j,] = pvalue
  }
}
if ((kernel=="RBF")|(kernel=="quadratic")){
  return (list("pvalue_KM"=pvalue_KM,"eigen.rate"=rate.vec))
}
if (kernel=="linear"){
  return (list("pvalue_KM"=pvalue_KM_original))
}
}

VTM<-function(vc, dm){
  matrix(vc, ncol=length(vc), nrow=dm, byrow=T)
}